@@ -25,7 +25,7 @@ from fondat.error import (
     NotFoundError,
 )
 from fondat.security import Scheme
-from fondat.stream import BytesStream, Stream, read_stream
+from fondat.stream import BytesStream, Reader, Stream
 from fondat.types import is_optional, is_subclass
 from typing import Annotated, Any, TypedDict
 
@@ -405,9 +405,13 @@ def get_param_in(method, param_name, type_hint):
                 return annotation()
             elif isinstance(annotation, (AsBody, InBody, InQuery)):
                 return annotation
-    if method._fondat_operation.type == "mutation":
-        return InBody(param_name)
-    return InQuery(param_name)
+    match method._fondat_operation.method:
+        case "post":
+            return InBody(param_name)
+        case "put":
+            return AsBody()
+        case _:
+            return InQuery(param_name)
 
 
 @functools.cache
@@ -432,10 +436,10 @@ def get_body_type(operation):
                     required_keys.add(param_in.name)
             case AsBody():
                 if as_body_param:
-                    raise TypeError("multiple AsBody annotated parameters")
+                    raise TypeError("multiple AsBody parameters")
                 as_body_param = name
     if as_body_param and in_body_params:
-        raise TypeError("mixed AsBody and InBody annotated parameters")
+        raise TypeError("mixed AsBody and InBody parameters")
     if as_body_param:
         return type_hints[as_body_param]
     if not in_body_params:
@@ -474,7 +478,7 @@ async def _decode_body(operation, request):
     if is_subclass(python_type, Stream):
         return request.body
     async with request.body as stream:  # close stream after reading
-        content = await read_stream(stream)
+        content = await Reader(stream).read()
     if len(content) == 0:
         return None  # empty body is no body
     try:
